#if !defined(_ARITHMETICCODEC_)
#define _ARITHMETICCODEC_


#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>

#define NUMBER_CHARACTERS	256	
#define EOF_SYMBOL			(NUMBER_CHARACTERS+1)
#define NUMBER_SYMBOLS		(NUMBER_CHARACTERS+1)

//Arrays fuer die Beziehungen zwischen Characters und Symbol-Indizes
long			arrayCharToIndex[NUMBER_CHARACTERS];
unsigned char	arrayIndexToChar[NUMBER_SYMBOLS+1];

long	freq[NUMBER_SYMBOLS+1];		//Haeufigkeiten der Symbole
long	cum_freq[NUMBER_SYMBOLS+1];	//wachsende Symbol-Haeufigkeiten



#define BITS_PER_CODE	16	//Anzahl der Bits fuer einen Code-Wert
typedef long intCodeValue;	//Typ eines Code-Wertes


//Groesster Code-Wert
#define CODE_TOP_VALUE (((long)1 << BITS_PER_CODE)-1)


//Definition der einzelnen Viertelgrenzen im Code-Wert-Intervall
#define FIRST_QUARTER	(CODE_TOP_VALUE/4 +1)
#define	HALF			(2*FIRST_QUARTER)
#define THIRD_QUARTER	(3*FIRST_QUARTER)



//Maximale Haeufigkeit
#define MAX_FREQUENCY	16383

//wachsende Symbol-Haeufigkeiten
extern long cum_freq[NUMBER_SYMBOLS+1];








//##################################################################
//##################################################################
class ArithmeticCodec
{
	public :
		FILE	*fin , *fout;

		long	buf;				//Bit-Buffer
		long	longBitsLeft;		//Anzahl der freien Bits in buf
		long	longGarbageBits;	//Anzahl der Fuellbits nach EOF

		//Unter- und Obergrenze des aktuellen Code-Wert-Intervalls
		intCodeValue	low , high;
		intCodeValue	value;

		//Anzahl der entgegengesetzten Bits , die nach dem
		//naechsten Bit auszugeben sind :
		long			longBitsFollowing;


		ArithmeticCodec(void){fin=NULL;fout=NULL;};
		ArithmeticCodec(FILE *f1 , FILE *f2)
		{
			fin = f1;
			fout = f2;
		};

		void	StartModel(void);
		void	UpdateModel(long symbol);

		void	StartEncoding(void);
		void	EncodeSymbol(long symbol ,long cum_freq[]);
		void	StopEncoding(void);

		void	StartDecoding(void);
		int		DecodeSymbol(long cum_freq[]);

		void	StartOutputingBits(void);
		void	OutputBit(int bit);
		void	StopOutputingBits(void);
		void	StartInputingBits(void);
		int		InputBit(void);
		void	OutputBitAndFollowing(int bit);

};//end of class






//#################################################################
void ArithmeticCodec::StartOutputingBits(void)
{
	buf = 0;
	longBitsLeft = 8;
}


//#################################################################
void ArithmeticCodec::OutputBit(int bit)
{
	buf >>= 1;
	//Das Bit an vorderste Stelle des buffers stellen
	if (bit)
		buf |= 0x80;
	longBitsLeft--;

	//Wenn der buffer voll ist , diesen schreiben
	if (longBitsLeft == 0)
	{
		putc((char)buf , fout);
		longBitsLeft = 8;
	}
}


//#################################################################
void ArithmeticCodec::StopOutputingBits(void)
{
	putc((char)(buf >> longBitsLeft) , fout);
}


//#################################################################
void ArithmeticCodec::StartInputingBits(void)
{
	longBitsLeft = 0;
	longGarbageBits = 0;
}


//#################################################################
int ArithmeticCodec::InputBit(void)
{
	int	ret;

	//Sind keine Bits mehr frei im buffer , dann das naechste Byte einlesen
	if (longBitsLeft == 0)
	{
		buf = getc(fin);

		//Nach EOF noch entsprechend viele willkuerliche Bits einfuegen
		if (buf == EOF)
		{
			longGarbageBits++;
			if (longGarbageBits > BITS_PER_CODE-2)
			{
				fprintf(stderr,"Bad input file\n");
				exit(-1);
			}
		}

		longBitsLeft = 8;
	}

	//Das naechste , unterste Bit des buffers zurueckliefern
	ret = buf & 1;
	buf >>= 1;
	longBitsLeft --;
	return ret;
}



//#################################################################
//Schreibt das bit und die evt. folgenden entgegengesetzten bits
void ArithmeticCodec::OutputBitAndFollowing(int bit)
{
	OutputBit(bit);
	while (longBitsFollowing > 0)
	{
		OutputBit(!bit);
		longBitsFollowing--;
	}
}





//##################################################################
//##################################################################
//Adaptives Modell initialisieren
void ArithmeticCodec::StartModel(void)
{
	int	i;

	//Beziehungen zwischen Characters und Symbol-Indizes aufbauen
	for (i=0; i<NUMBER_CHARACTERS; i++)
	{
		arrayCharToIndex[i] = i+1;
		arrayIndexToChar[i+1] = (unsigned char) i;
	}

	//Anfangshaeufigkeiten initialisieren (1 fuer alle Symbole)
	for (i=0; i<=NUMBER_SYMBOLS; i++)
	{
		freq[i] = 1;
		cum_freq[i] = NUMBER_SYMBOLS-i;
	}

	//freq[0] muss verschieden von freq[1] sein
	freq[0] = 0;
}


//##################################################################
//##################################################################
//Modell fuer ein neu aufgetretenes Symbol aktualisieren
void ArithmeticCodec::UpdateModel(long symbol)
{
	int newindex;

	//Maximalanzahl an Haeufigkeiten erreicht ?
	//Wenn ja , dann alle halbieren (aber ungleich 0 halten)
	if (cum_freq[0] == MAX_FREQUENCY)
	{
		int cum;
		cum = 0;
		for (newindex=NUMBER_SYMBOLS; newindex>=0; newindex--)
		{
			freq[newindex] = (freq[newindex]+1)/2;
			cum_freq[newindex] = cum;
			cum += freq[newindex];
		}
	}
    
	//Den neuen Index fuer das Symbol suchen
	for (newindex=symbol; freq[newindex]==freq[newindex-1]; newindex--) ;


	//Falls noetig , Beziehungstabellen aktualisieren
    if (newindex < symbol)
	{
		int ch_i , ch_symbol;
		ch_i = arrayIndexToChar[newindex];
		ch_symbol = arrayIndexToChar[symbol];
		arrayIndexToChar[newindex] = (unsigned char) ch_symbol;
		arrayIndexToChar[symbol] = (unsigned char) ch_i;
		arrayCharToIndex[ch_i] = symbol;
		arrayCharToIndex[ch_symbol] = newindex;
	}

	//Haeufigkeit erhoehen
	freq[newindex]++;

	//Haeufigkeiten aktualisieren
	while (newindex > 0)
	{
		newindex--;
		cum_freq[newindex]++;
	}
}




//##################################################################
//##################################################################
void ArithmeticCodec::StartEncoding(void)
{
	low = 0;
	high = CODE_TOP_VALUE;
	longBitsFollowing = 0;
}



//##################################################################
//##################################################################
void ArithmeticCodec::EncodeSymbol(long symbol,long cum_freq[] )
{
	long range;		//Groesse des aktuelle Code-Wert-Intervalls
	range = (long)(high-low)+1;

	//Intervallgrenzen auf die diesem Symbol zugewiesenen
	//Grenzen verkleinern
	high = low +
		(range*cum_freq[symbol-1])/cum_freq[0]-1;
	low = low +
		(range*cum_freq[symbol])/cum_freq[0];

	for (;;)
	{
		//befinden wir uns in der unteren Haelfte , dann 0 ausgeben
		if (high < HALF)
		{
			OutputBitAndFollowing(0);
		}
		//und fuer die obere Haelfte eine 1 ausgeben
		else if (low >= HALF)
		{
			OutputBitAndFollowing(1);
			//Offset zur Obergrenze subtrahieren
			low -= HALF;
			high -= HALF;
		}
		//Wenn in der mittleren Haelfte , dann spaeter ein
		//entgegengesetztes Bit ausgeben
		else if ( (low >= FIRST_QUARTER) && (high < THIRD_QUARTER) )
		{
			longBitsFollowing++;
			//Offset zur Mitte subtrahieren
			low -= FIRST_QUARTER;
			high -= FIRST_QUARTER;
		}
		else break;

		//Code-Wert-Intervall skalieren
		low = 2*low;
		high = 2*high+1;
	}
}



//##################################################################
//##################################################################
void ArithmeticCodec::StopEncoding(void)
{
	//zwei Bits ausgeben , welche demjenigen Viertel entsprechen ,
	//in dem man sich gerade befindet
	longBitsFollowing++;
	if (low < FIRST_QUARTER)
		OutputBitAndFollowing(0);
	else OutputBitAndFollowing(1);
}





//##################################################################
//##################################################################
void ArithmeticCodec::StartDecoding(void)
{
	int i;
	value = 0;
	for (i=1; i<=BITS_PER_CODE; i++)
	{
		value = 2*value+InputBit();
	}

	//Code-Wert-Intervall initialisieren
	low = 0;
	high = CODE_TOP_VALUE;
}



//##################################################################
//##################################################################
int ArithmeticCodec::DecodeSymbol(long cum_freq[])
{
	long range;		//Groesse des aktuellen Intervalls
	int cum;		//berechnete anwachsende Haeufigkeit
	int symbol;

	range = (long)(high-low)+1;

	//Fuer den Code-Wert die Haeufigket suchen
	cum = (int)
		((((long)(value-low)+1)*cum_freq[0]-1)/range);

	//Und dann das Symbol selber
	for (symbol = 1; cum_freq[symbol]>cum; symbol++);

	//Die Code-Wert-Intervall Grenzen auf die dem Symbol zugewiesenen
	//Grenzen einengen
	high = low +
		(range*cum_freq[symbol-1])/cum_freq[0]-1;
	low = low +
		(range*cum_freq[symbol])/cum_freq[0];
	
	for (;;)
	{
		//untere Haelfte des Intervalls ausweiten
		if (high < HALF)
		{
		}
		//obere Haelfte ausweiten
		else if (low >= HALF)
		{
			value -= HALF;
			low -= HALF;	//Offset zur Obergrenze subtrahieren
			high -= HALF;
		}
		//mittlere Haelfte ausweiten
		else if ( (low >= FIRST_QUARTER) && (high < THIRD_QUARTER) )
		{
			value -= FIRST_QUARTER;
			low -= FIRST_QUARTER;	//offset zur Mitte subtrahieren
			high -= FIRST_QUARTER;
		}
		else break;

		//Code-Wert-Intervall skalieren
		low = 2*low;
		high = 2*high+1;

		//naechstes InputBit holen
		value = 2*value+InputBit();
	}

	return symbol;
}





#endif